from json_cpp import JsonObject, JsonList
from .util import *
from .coordinates import *
from .location import *
from .shape import Space
from .world import World_implementation, World
from .cell import Cell_group_builder
from datetime import datetime


class Step(JsonObject):
    def __init__(self,
                 time_stamp: float = 0.0,
                 agent_name: str = "",
                 frame: int = 0,
                 location: Location = None,
                 rotation: float = 0.0,
                 data: str = ""):
        self.time_stamp = time_stamp
        self.agent_name = agent_name
        self.frame = frame
        if location is None:
            location = Location(0, 0)
        self.location = location
        self.rotation = rotation
        self.data = data


class Velocities(JsonList):
    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable, list_type=float)

    def complementary_filter(self, a: float): #complementary filter
        if a <= 0 or a >= 1:
            raise ArithmeticError("filter parameter should be > 0 and < 1")
        filtered = Velocities()
        last = self[0]
        for v in self:
            nv = (a * last + (1 - a) * v)
            last = nv
            filtered.append(nv)
        return filtered

    def outliers_filter(self, threshold: float):
        if threshold <= 0:
            raise ArithmeticError("threshold parameter should be > 0")
        threshold = sum(self)/len(self) * threshold
        filtered = Velocities()
        if len(self) == 0:
            return filtered
        last = self[0]
        for v in self:
            if abs(v-last) <= threshold:
                filtered.append(v)
                last = v
            else:
                filtered.append(last)
        return filtered

    def moving_average_filter(self, window_length: int = 2):
        if window_length <= 0:
            raise ArithmeticError("window length parameter should be > 0")
        filtered = Velocities()
        if len(self) == 0:
            return filtered
        le = len(self) - window_length
        for i in range(len(self)):
            b = i - window_length if i > window_length else 0
            e = i + window_length if i < le else len(self)
            filtered.append(sum(self[b:e])/(e-b))
        return filtered


class Trajectories(JsonList):
    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable, list_type=Step)

    def filter_initial_bad_readings(self):
        first_frame = 1
        for i, step in enumerate(self):
            if step.frame != 0:
                break
            first_frame = i
        del self[0:first_frame-1]

    def get_step_index_by_time_stamp(self, time_stamp: float, exact: bool = False) -> int:
        if len(self) == 1:
            if self[0].time_stamp != time_stamp and exact:
                raise RuntimeError("Time_stamp %i not found", time_stamp)
            else:
                return 0
        m = len(self) // 2
        if self[m].time_stamp == time_stamp:
            return m
        if time_stamp > self[m].time_stamp:
            return m + Trajectories.get_step_index_by_time_stamp(self[m:], time_stamp, exact)
        else:
            return Trajectories.get_step_index_by_time_stamp(self[:m], time_stamp, exact)

    def get_step_by_time_stamp(self, time_stamp: float, exact: bool = False) -> Step:
        step_index = self.get_step_index_by_time_stamp(time_stamp, exact)
        return self[step_index]

    def get_step_index_by_frame(self, frame: int, exact: bool = True) -> int:
        if len(self) == 1:
            if self[0].frame != frame and exact:
                raise RuntimeError("Frame %i not found", frame)
            else:
                return 0
        m = len(self) // 2
        if self[m].frame == frame:
            return m
        if frame > self[m].frame:
            return m + Trajectories.get_step_index_by_frame(self[m:], frame, exact)
        else:
            return Trajectories.get_step_index_by_frame(self[:m], frame, exact)

    def get_step_by_frame(self, frame: int, exact: bool = True) -> Step:
        try:
            step_index = self.get_step_index_by_frame(frame, exact)
            return self[step_index]
        except:
            return None

    def get_velocities(self) -> {}:
        velocities = {}
        last_locations = {}
        last_time_stamp = {}
        for s in self:
            if s.agent_name not in velocities:
                velocities[s.agent_name] = Velocities()
                velocities[s.agent_name].append(0.0)
            else:
                time_dif = s.time_stamp - last_time_stamp[s.agent_name]
                velocity = s.location.dist(last_locations[s.agent_name]) / time_dif
                velocities[s.agent_name].append(velocity)
            last_locations[s.agent_name] = s.location
            last_time_stamp[s.agent_name] = s.time_stamp
        return velocities

    def get_stops_steps(self, distance_threshold: float = 0.01, stop_time: float = 0.5):
        stops = []
        future = 0
        for current in range(len(self)):
            if current <= future:
                continue
            future = current
            last_future = future
            while future < len(self) and \
                    self[current].location.dist(self[future].location) < distance_threshold:
                #in case of a big break between frames
                if self[future].time_stamp - self[last_future].time_stamp > stop_time:
                    #it repositions itself in the frame before the break and halts
                    future = last_future
                    break
                last_future = future
                future += 1
            if future == len(self):
                break
            if self[current].time_stamp + stop_time < self[future].time_stamp:
                stops.append((self[current], self[future]))
            else:
                future = 0
        return stops

    def get_stops(self, distance_threshold: float = 0.01, stop_time: float = 0.5):
        return [(b.frame, e.frame) for b, e in self.get_stops_steps(distance_threshold, stop_time)]

    def get_filtered_velocities(self, complementary: float = None, outliers: float = None) -> {}:
        avs = self.get_velocities()
        for agent_name in avs:
            if outliers:
                avs[agent_name] = avs[agent_name].outliers_filter(outliers)
            if complementary:
                avs[agent_name] = avs[agent_name].complementary_filter(complementary)
        return avs

    def get_unique_steps(self):
        unique_steps = Trajectories()
        last_locations = {}
        for s in self:
            if s.agent_name not in last_locations or not last_locations[s.agent_name] == s.location:
                unique_steps.append(s)
            last_locations[s.agent_name] = s.location
        return unique_steps

    def get_agent_names(self) -> list:
        agent_names = []
        for s in self:
            if s.agent_name not in agent_names:
                agent_names.append(s.agent_name)
        return agent_names

    def get_agent_trajectory(self, agent_name: str):
        return self.where("agent_name", agent_name)


class Episode(JsonObject):
    def __init__(self,
                 start_time: datetime = None,
                 time_stamp: float = 0.0,
                 end_time: datetime = None,
                 trajectories: Trajectories = None,
                 captures: JsonList = None,
                 rewards_sequence: Cell_group_builder = None,
                 rewards_time_stamps: JsonList = None):
        if not start_time:
            start_time = datetime.now()
        self.start_time = start_time
        self.time_stamp = time_stamp
        if not end_time:
            end_time = datetime.now()
        self.end_time = end_time
        if trajectories is None:
            trajectories = Trajectories()
        self.trajectories = trajectories
        if captures is None:
            captures = JsonList(list_type=int)
        self.captures = captures
        if rewards_sequence:
            self.rewards_sequence = rewards_sequence
        else:
            self.rewards_sequence = Cell_group_builder()
        if rewards_time_stamps:
            self.rewards_time_stamps = rewards_time_stamps
        else:
            self.rewards_time_stamps = JsonList(list_type=float)

    def clean(self):
        ts = self.trajectories.get_agent_trajectory("predator").where("frame", 0).get("time_stamp")
        if len(ts) > 0:
            min_time_stamp = min(ts)
            new_trajectories = Trajectories()
            for step in self.trajectories:
                if step.frame != 0 or step.agent_name != "predator" or step.time_stamp == min_time_stamp:
                    new_trajectories.append(step)
            self.trajectories = new_trajectories


class Episode_list(JsonList):
    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable, list_type=Episode)


class Experiment(JsonObject):
    def __init__(self,
                 name: str = "",
                 world_configuration_name: str = "",
                 world_implementation_name: str = "",
                 occlusions: str = "",
                 subject_name: str = "",
                 duration: int = 0,
                 start_time: datetime = None,
                 episodes: Episode_list = None,
                 episode_count: int = 0,
                 rewards_cells: Cell_group_builder = [],
                 rewards_orientations: JsonList = []):
        self.name = name
        self.world_configuration_name = world_configuration_name
        self.world_implementation_name = world_implementation_name
        self.occlusions = occlusions
        self.subject_name = subject_name
        self.duration = duration
        if start_time is None:
            start_time = datetime.now()
        self.start_time = start_time
        if episodes is None:
            episodes = Episode_list()
        self.episodes = episodes
        self.episode_count = episode_count
        if rewards_cells:
            self.rewards_cells = rewards_cells
        else:
            self.rewards_cells = Cell_group_builder()
        if rewards_orientations:
            self.rewards_orientations = rewards_orientations
        else:
            self.rewards_orientations = JsonList(list_type=int)

    def clean_all_episodes (self):
        for episode in self.episodes:
            episode.clean()

    @staticmethod
    def get_from_file(file_name: str):
        e = Experiment.parse(open(file_name).read())
        check_type(e, Experiment, "")
        return e

    @staticmethod
    def get_from_url(url: str):
        e = Experiment.parse(get_web_json(resource_uri=url))
        check_type(e, Experiment, "")
        return e

    def transform(self, dst_space_name: str):
        check_type(dst_space_name, str, "incorrect type for dst_space")

        dst_space = World_implementation.get_from_name(self.world_configuration_name, dst_space_name).space

        src_space = World_implementation.get_from_name(self.world_configuration_name, self.world_implementation_name).space
        for episode in self.episodes:
            for step in episode.trajectories:
                step.location = Space.transform_to(step.location, src_space, dst_space)
        self.world_implementation_name = dst_space_name

    def set_name(self, prefix: str = None, suffix: str = None):
        self.name = self.start_time.strftime("%Y%m%d_%OH%M_") + self.subject_name + "_" + self.occlusions
        if prefix:
            self.name = prefix + "_" + self.name
        if suffix:
            self.name = self.name + "_" + suffix

    def get_wrong_origin_episodes(self, origin: Coordinates = Coordinates(-20, 0), agent_name: str = "prey") -> JsonList:
        world = World.get_from_parameters_names(self.world_configuration_name, "canonical")
        wrong_origin_episodes = JsonList(list_type=int)
        for i, episode in enumerate(self.episodes):
            agent_trajectory = episode.trajectories.get_agent_trajectory(agent_name)
            if len(agent_trajectory) == 0:
                continue
            episode_origin_location = agent_trajectory[0].location
            episode_origin_index = world.cells.find(episode_origin_location)
            episode_origin_coordinates = world.cells[episode_origin_index].coordinates
            if not episode_origin_coordinates == origin:
                wrong_origin_episodes.append(i)
        return wrong_origin_episodes

    def get_wrong_goal_episodes(self, goal: Coordinates = Coordinates(20, 0), agent_name: str = "prey") -> JsonList:
        world = World.get_from_parameters_names(self.world_configuration_name, "canonical")
        wrong_goal_episodes = JsonList(list_type=int)
        for i, episode in enumerate(self.episodes):
            agent_trajectory = episode.trajectories.get_agent_trajectory(agent_name)
            if len(agent_trajectory) == 0:
                continue
            episode_goal_location = agent_trajectory[-1].location
            episode_goal_index = world.cells.find(episode_goal_location)
            episode_goal_coordinates = world.cells[episode_goal_index].coordinates
            if not episode_goal_coordinates == goal:
                wrong_goal_episodes.append(i)
        return wrong_goal_episodes

    def get_incomplete_episodes(self, threshold_step_count: int = 30, agent_name: str = "prey") -> JsonList:
        incomplete_episodes = JsonList(list_type=int)
        for i, episode in enumerate(self.episodes):
            agent_trajectory = episode.trajectories.get_agent_trajectory(agent_name)
            if len(agent_trajectory) == 0:
                continue
            if len(agent_trajectory) < threshold_step_count:
                incomplete_episodes.append(i)
        return incomplete_episodes

    def get_broken_trajectory_episodes(self, threshold_distance: float = .1, agent_name: str = "prey") -> JsonList:
        broken_trajectory_episodes = JsonList(list_type=int)
        for i, episode in enumerate(self.episodes):
            agent_trajectory = episode.trajectories.get_agent_trajectory(agent_name)
            last_location = None
            for step in agent_trajectory:
                if last_location is not None:
                    if last_location.dist(step.location) > threshold_distance:
                        broken_trajectory_episodes.append(i)
                        break
                last_location = step.location
        return broken_trajectory_episodes

    def remove_episodes(self, episode_list: JsonList) -> None:
        for episode_index in episode_list:
            self.episodes[episode_index].trajectories = Trajectories()
            self.episodes[episode_index].captures = JsonList(list_type=int)



from .location import Location
from .shape import Polygon_list, Polygon
from .cell import Cell_group
from .util import to_radians
from .util import angle_difference


class Location_visibility:

    def __init__(self, occlusions: Polygon_list):
        self.occlusions = occlusions

    def is_visible(self, src: Location, dst: Location) -> bool:
        theta = src.atan(dst)
        dist = src.dist(dst)
        for occlusion in self.occlusions:
            if occlusion.is_between(src, theta=theta, dist=dist):
                return False
        return True

    def visible_cells(self, src: Location, cells: Cell_group) -> Cell_group:
        """
        Finds all visible cells from specified location
        :return
        A list of all the visible cell ids
        """
        free_cells = cells.free_cells()
        visible_cells = Cell_group()

        for free_cell in free_cells:
            if self.is_visible(src, free_cell.location):
                visible_cells.append(free_cell)

        return visible_cells

    def hidden_cells(self, src: Location, cells: Cell_group) -> Cell_group:
        """
        Finds all hidden cells from specified location arg (loc)
        :return:
        A list of all the hidden cell ids
        """
        hidden_cells = Cell_group()
        free_cells = cells.free_cells()
        visible_cells_ids = self.visible_cells(src, cells).get('id')
        for cell in free_cells:
            if cell.id not in visible_cells_ids:
                hidden_cells.append(cell)

        return hidden_cells






import numpy
from matplotlib.patches import RegularPolygon
import matplotlib.pyplot as plt
import matplotlib.colors
from matplotlib.path import Path
from matplotlib.transforms import Affine2D
import matplotlib.colors as mcolors
from .world import *
from .experiment import *
from .agent_markers import *
from .QuickBundles import *


def adjust_color_brightness(color, amount:float=1):
    import matplotlib.colors as mc
    import colorsys
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])


class Display:

    def __init__(self,
                 world: World,
                 fig_size: tuple = (10, 10),
                 padding: float = .1,
                 outline: float = .5,
                 show_axes: bool = False,
                 cell_color="white",
                 occlusion_color="black",
                 background_color="white",
                 habitat_color="white",
                 cell_edge_color="lightgray",
                 cell_fill: bool = True,
                 show_cell: bool = True,
                 show_occluded_cell: bool = True,
                 show_habitat: bool = True,
                 habitat_fill: bool = True,
                 occluded_cell_edge_color="lightgray",
                 habitat_edge_color="gray",
                 cell_outline_zorder: int = 2,
                 cell_zorder: int = 3,
                 occluded_cell_zorder: int = 3,
                 habitat_zorder: int = 1,
                 cell_alpha: int = 1,
                 cell_outline_alpha: int = 1,
                 habitat_alpha: int = 1,
                 animated: bool = False,
                 ax=None,
                 fig=None):
        if animated:
            plt.ion()
        self.cell_alpha = cell_alpha
        self.show_cell = show_cell
        self.show_occluded_cell = show_occluded_cell
        self.show_habitat = show_habitat
        self.cell_outline_alpha = cell_outline_alpha
        self.habitat_alpha = habitat_alpha
        self.cell_fill = cell_fill
        self.habitat_fill = habitat_fill
        self.agents = dict()
        self.agents_markers = dict()
        self.agents_markers["predator"] = Agent_markers.robot()
        self.agents_markers["prey"] = Agent_markers.mouse()
        self.animated = animated
        self.world = world
        if not fig:
            self.fig = plt.figure(figsize=fig_size)
        else:
            self.fig = fig
        if not ax:
            self.ax = self.fig.add_subplot(111)
        else:
            self.ax = ax
        self.ax.axes.xaxis.set_visible(show_axes)
        self.ax.axes.yaxis.set_visible(show_axes)
        self.agents_trajectories = Trajectories()
        self.agents_markers = dict()
        self.cells = []
        self.occlusion_color = occlusion_color
        self.habitat_color = habitat_color
        self.habitat_edge_color = habitat_edge_color
        self.cell_color = cell_color
        self.cell_edge_color = cell_edge_color
        self.occluded_cell_edge_color = occluded_cell_edge_color
        self.xcenter = world.implementation.space.center.x
        self.ycenter = world.implementation.space.center.y
        self.outline = outline
        hsize = world.implementation.space.transformation.size / 2
        pad = hsize * padding

        xmin = self.xcenter - hsize - pad
        xmax = self.xcenter + hsize + pad

        ymin = self.ycenter - hsize - pad
        ymax = self.ycenter + hsize + pad

        self.ax.set_xlim(xmin=xmin, xmax=xmax)
        self.ax.set_ylim(ymin=ymin, ymax=ymax)
        self.ax.set_facecolor(background_color)
        self.habitat_theta = math.radians(0 - world.implementation.space.transformation.rotation)
        self.cells_theta = math.radians(0 - world.implementation.cell_transformation.rotation) + self.habitat_theta
        self.cells_size = world.implementation.cell_transformation.size / 2
        self.habitat_size = hsize
        self.cell_polygons = []
        self.cell_outline_polygons = []
        self.habitat_polygon = None
        self.background_extent = [self.xcenter - hsize, self.xcenter + hsize, self.ycenter - hsize, self.ycenter + hsize]
        self.cell_outline_zorder = cell_outline_zorder
        self.cell_zorder = cell_zorder
        self.occluded_cell_zorder = occluded_cell_zorder
        self.habitat_zorder = habitat_zorder
        self._draw_cells__()
        plt.tight_layout()

    def _draw_cells__(self):
        [p.remove() for p in reversed(self.ax.patches)]
        for cell in self.world.cells:
            color = self.occlusion_color if cell.occluded else self.cell_color
            edge_color = self.occluded_cell_edge_color if cell.occluded else self.cell_edge_color
            fill = True if cell.occluded else self.cell_fill
            visible = self.show_occluded_cell if cell.occluded else self.show_cell
            self.cell_outline_polygons.append(self.ax.add_patch(RegularPolygon((cell.location.x, cell.location.y),
                                                                               self.world.configuration.cell_shape.sides,
                                                                               self.cells_size,
                                                                               facecolor=color,
                                                                               edgecolor=edge_color,
                                                                               orientation=self.cells_theta,
                                                                               zorder=self.cell_outline_zorder,
                                                                               linewidth=1,
                                                                               alpha=self.cell_outline_alpha,
                                                                               fill=fill,
                                                                               visible=visible)))
            self.cell_polygons.append(self.ax.add_patch(RegularPolygon((cell.location.x, cell.location.y),
                                                                       self.world.configuration.cell_shape.sides,
                                                                       self.cells_size * self.outline,
                                                                       facecolor=color,
                                                                       orientation=self.cells_theta,
                                                                       zorder=self.cell_zorder,
                                                                       linewidth=0,
                                                                       alpha=self.cell_alpha,
                                                                       fill=fill,
                                                                       visible=visible)))
        self.habitat_polygon = self.ax.add_patch(RegularPolygon((self.xcenter, self.ycenter),
                                                                self.world.implementation.space.shape.sides,
                                                                self.habitat_size,
                                                                facecolor=self.habitat_color,
                                                                edgecolor=self.habitat_edge_color,
                                                                orientation=self.habitat_theta,
                                                                zorder=self.habitat_zorder,
                                                                alpha=self.habitat_alpha,
                                                                fill=self.habitat_fill,
                                                                visible=self.show_habitat))

    def set_occlusions(self, occlusions: Cell_group_builder):
        self.world.set_occlusions(occlusions)
        self._draw_cells__()

    def set_agent_marker(self, agent_name: str, marker: Path):
        self.agents_markers[agent_name] = marker

    def add_trajectories(self, trajectories: Trajectories, colors={}, alphas={}, start_frame=0, end_frame=None, start_time=0, end_time=None, zorder=4, auto_alpha=False):
        agents = trajectories.get_agent_names()
        for index, agent in enumerate(agents):
            agent_trajectory = trajectories.get_agent_trajectory(agent)
            x = []
            y = []
            locations=Location_list()
            for step in agent_trajectory:
                if step.frame < start_frame or (end_frame and step.frame > end_frame):
                    continue
                if step.time_stamp < start_time or (end_time and step.time_stamp > end_time):
                    continue
                x.append(step.location.x)
                y.append(step.location.y)
                locations.append(step.location)
            color = list(matplotlib.colors.cnames.keys())[index]
            alpha = 0.5
            if agent in colors:
                color = colors[agent]
            if agent in alphas:
                alpha = alphas[agent]
            if auto_alpha:
                dist = locations.distance()
                if not dist or dist[-1] == 0:
                    alpha = 1
                else:
                    alpha = [d/dist[-1]/2 for d in dist]

            for i in range(len(x)-1):
                lcolor = None
                lalpha = None
                if type(alpha) is list:
                    lalpha = alpha[i]
                else:
                    lalpha = alpha
                if type(color) is numpy.ndarray:
                    lcolor = color[i]
                else:
                    lcolor = color
                self.ax.plot([x[i], x[i+1]], [y[i], y[i+1]], color=lcolor, alpha=lalpha, linewidth=3, zorder=zorder)


    def cell(self, cell: Cell = None, cell_id: int = -1, coordinates: Coordinates = None, color=None, outline_color=None, edge_color=None, alpha=None):
        if color is None:
            color = self.cell_color
        if edge_color is None:
            edge_color = self.cell_edge_color
        if cell is None:
            if cell_id == -1:
                if coordinates is None:
                    raise RuntimeError("a cell, cell_id or coordinates must be provided")
                for c in self.world.cells:
                    if c.coordinates == coordinates:
                        cell = c
                        break
                if cell is None:
                    raise RuntimeError("cell coordinates not found")
            else:
                cell = self.world.cells[cell_id]
        if outline_color is None:
            outline_color = color
        if alpha is None:
            alpha = self.cell_alpha
        self.cell_polygons[cell.id].set_facecolor(color)
        self.cell_outline_polygons[cell.id].set_edgecolor(edge_color)
        self.cell_outline_polygons[cell.id].set_facecolor(outline_color)
        self.cell_outline_polygons[cell.id].set_alpha(alpha)

    def heatmap(self, values: list, color_map=plt.cm.Reds, value_range: tuple = None) -> None:
        if value_range:
            minv, maxv = value_range
        else:
            minv, maxv = min(values), max(values)

        if (minv == maxv):
            return
        adjusted_values = [(v-minv)/(maxv-minv) for v in values]
        adjusted_cmap = color_map(adjusted_values)
        for cell_id, color in enumerate(adjusted_cmap):
            if not self.world.cells[cell_id].occluded:
                self.cell(cell_id=cell_id, color=color)

    def circle(self, location: Location, radius: float, color, alpha: float = 1.0, zorder: int = 4):
        circle_patch = plt.Circle((location.x, location.y), radius, color=color, alpha=alpha, zorder=zorder)
        return self.ax.add_patch(circle_patch)

    def arrow(self, beginning: Location, ending: Location = None, theta: float = 0, dist: float = 0, color="black", head_width: float = .02, alpha: float = 1.0, line_width: float = 0.001, zorder: int = 4, existing_arrow: matplotlib.patches.FancyArrowPatch = None) -> matplotlib.patches.FancyArrowPatch:
        if ending is None:
            ending = beginning.copy().move(theta=theta, dist=dist)
        length = ending - beginning
        if existing_arrow:
            existing_arrow.set_data(x=beginning.x, y=beginning.y, dx=length.x, dy=length.y, head_width=head_width, width=line_width)
            existing_arrow.set_color(color)
            existing_arrow.set_alpha(alpha)
            return existing_arrow
        else:
            new_arrow = self.ax.arrow(beginning.x, beginning.y, length.x, length.y, color=color, head_width=head_width, length_includes_head=True, alpha=alpha, width=line_width, zorder=zorder)
            new_arrow.animated = self.animated
            return new_arrow

    def line(self, beginning: Location, ending: Location = None, theta: float = 0, dist: float = 0, color="black", alpha: float = 1.0, line_width: float = 0.001, zorder: int = 4, existing_line: matplotlib.patches.FancyArrowPatch = None) -> matplotlib.patches.FancyArrowPatch:
        head_width = 0
        if ending is None:
            ending = beginning.copy().move(theta=theta, dist=dist)
        length = ending - beginning
        if existing_line:
            existing_line.set_data(x=beginning.x, y=beginning.y, dx=length.x, dy=length.y, head_width=0, head_length=0, alpha=alpha, width=line_width)
            existing_line.set_color(color)
            existing_line.set_alpha(alpha)
            return existing_line
        else:
            new_arrow = self.ax.arrow(beginning.x, beginning.y, length.x, length.y, color=color, head_width=0, alpha=alpha, head_length=0, length_includes_head=False, width=line_width, zorder=zorder)
            new_arrow.animated = self.animated
            return new_arrow
    def set_background(self, background, alpha:float = 1):
        self.ax.imshow(background, extent=self.background_extent, alpha=alpha)
        self._draw_cells__()

    def agent(self, step: Step = None, agent_name: str = None, location: Location = None, rotation: float = None, color=None, size: float = 40.0, show_trajectory: bool = True, zorder: int = 4, marker: Path=None):
        if step:
            agent_name = step.agent_name
            location = step.location
            rotation = step.rotation

        # if show_trajectory:
        #     self.agents_trajectories.append(step)
        #     x = self.agents_trajectories.get_agent_trajectory(agent_name).get("location").get("x")
        #     y = self.agents_trajectories.get_agent_trajectory(agent_name).get("location").get("y")
        #     self.agents[agent_name], = self.ax.plot(x, y, c=color)

        if not marker:
            if agent_name in self.agents_markers:
                marker = self.agents_markers[agent_name]
            else:
                if agent_name == "predator":
                    marker = Agent_markers.robot()
                else:
                    marker = Agent_markers.mouse()

        if agent_name not in self.agents:
            self.agents[agent_name], = self.ax.plot(location.x, location.y, marker=marker, c=color, markersize=size, zorder=zorder)

        t = Affine2D().rotate_deg_around(0, 0, -rotation)
        self.agents[agent_name].set_marker(marker.transformed(t))
        self.agents[agent_name].set_xdata(location.x)
        self.agents[agent_name].set_ydata(location.y)
        self.agents[agent_name].set_color(color)

    def plot_clusters(self, clusters, colors: list=["blue", "red"], show_centroids: bool = True, use_alpha: bool = True, zorder: int = 4):
        for sl in clusters.unclustered:
            self.ax.plot([s.x for s in sl], [s.y for s in sl], c="gray", linewidth=2, alpha=.5)

        for cn, cluster in enumerate(clusters):
            distances = cluster.get_distances()
            max_distance = max(distances)
            if use_alpha:
                alpha_values = [1 if max_distance == 0 else (max_distance - d) / max_distance / 2 for d in distances]
            else:
                alpha_values = [1 for d in distances]
            color = adjust_color_brightness(colors[cn % len(colors)], .8)
            for i, sl in enumerate(cluster):
                self.ax.plot([s.x for s in sl], [s.y for s in sl], c=color, linewidth=2, alpha=alpha_values[i], zorder=zorder)

        if show_centroids:
            for cn, cluster in enumerate(clusters):
                color = adjust_color_brightness(colors[cn % len(colors)], 1.2)
                cluster_size = .005 + .01 * (len(cluster) / clusters.streamline_count())
                for s in cluster.centroid:
                    self.circle(s, radius=cluster_size + .003, color="lightgray", alpha=1, zorder=500)
                for s in cluster.centroid:
                    self.circle(s, radius=cluster_size, color=color, alpha=1, zorder=500)

    def update(self):
        if self.animated:
            self.fig.canvas.draw()
            self.fig.canvas.flush_events()
            plt.pause(.001)


def LoadVideo(video_path: str) -> list:
    import cv2
    with cv2.VideoCapture(video_path) as video_reader:
        frames = []
        while video_reader.isOpened():
            ret, video_frame = video_reader.read()
            if video_frame is None:
                break
            frames.append(video_frame)
    return frames


def CropVideo(frames:list, crop=(-980,None,None,None,None,None)) -> list:
    cropped_frames = []
    for frame in frames:
        cropped_frames.append(frame[crop[0]:crop[1], crop[2]:crop[3], crop[4]:crop[5]])
    return cropped_frames

def EpisodeRepresentation(episode: Episode, world: World, frames: list=None, time_stamps: list=None, colors={"prey": "blue", "predator": "red"}, cols=0, figsize=(10,10), video_path:str="", video_frames:list=None, **kwargs):

    if video_path:
        video_frames = CropVideo(LoadVideo(video_path))
    else:
        video_frames = None

    selections = frames if frames else time_stamps
    if cols == 0:
        cols = len(selections)
    rows = math.ceil(len(selections)/cols)
    fig = plt.figure(figsize=figsize)
    axs = fig.subplots(rows, cols)
    fig.tight_layout()
    plt.subplots_adjust(wspace=0, hspace=0)
    ps = 0
    displays = []
    if frames:
        selection_frames = frames
    else:
        selection_frames = []
        pf = 0
        for ts in time_stamps:
            for i, step in enumerate(episode.trajectories[pf:]):
                if step.time_stamp > ts:
                    selection_frames.append(step.frame)
                    break
                pf = i
            else:
                selection_frames.append(step.frame)
            print (selection_frames)

    for i, selection in enumerate(selections):
        ax = axs[int(i/cols),i % cols]
        if video_frames:
            displays.append(Display(world, cell_fill=False, background_color="black", habitat_fill=False, show_cell=False, show_occluded_cell=False, ax=ax))
            displays[-1].background_extent = [0, 1, .0435, .95]
            displays[-1].set_background(video_frames[selection_frames[i]], 1)
        else:
            displays.append(Display(world, cell_fill=True, background_color="black", habitat_fill=True, show_cell=True, show_occluded_cell=True, ax=ax))

        if frames:
            displays[-1].add_trajectories(episode.trajectories, colors=colors, start_frame=ps, end_frame=selection, auto_alpha=True)
        else:
            displays[-1].add_trajectories(episode.trajectories, colors=colors, start_time=ps, end_time=selection, auto_alpha=True)
        ps = selection
    return fig, displays


import math
import requests
import json
global cellworld_data_base_uri
from datetime import datetime, timedelta
import os

cellworld_data_base_uri = "https://raw.githubusercontent.com/germanespinosa/cellworld_data/master/"
cellworld_cache_folder = os.environ.get("CELLWORLD_CACHE")


def get_resource(resource_type: str, key0, *argv):
    resource_uri = cellworld_data_base_uri + resource_type + "/" + key0
    file_path = ""
    if cellworld_cache_folder:
        file_path = cellworld_cache_folder + "/" + resource_type + "/" + key0
    for arg in argv:
        resource_uri += "." + arg
        if cellworld_cache_folder:
            file_path += "." + arg
    if cellworld_cache_folder and os.path.exists(file_path):
        content = ""
        with open(file_path, "r") as text_file:
            content = text_file.read()
        return json.loads(content)
    else:
        response = requests.get(resource_uri)
        return json.loads(response.text)


def get_web_json(resource_uri: str):
    response = requests.get(resource_uri)
    return json.loads(response.text)


def to_radians(rotation: float) -> float:
    return rotation / 180 * math.pi


def to_degrees(theta: float) -> float:
    return theta * 180 / math.pi


class Time_out:
    def __init__(self, seconds=1.0):
        self.end_time = datetime.now() + timedelta(seconds=seconds)

    def __bool__(self):
        return self.end_time > datetime.now()


class Timer:
    def __init__(self, seconds: float = 0):
        self.time = seconds
        self.check_point = datetime.now()

    def to_seconds(self) -> float:
        return (datetime.now() - self.check_point).total_seconds()

    def __bool__(self):
        return self.to_seconds() < self.time

    def time_out(self) -> bool:
        return self.to_seconds() > self.time

    def reset(self):
        self.check_point = datetime.now()


def check_type(v, t, m):
    if not isinstance(v, t):
        raise TypeError(m)


def check_types(v, ts, m):
    r = False
    for t in ts:
        if isinstance(v, t):
            r = True
    if not r:
        raise TypeError(m)


def normalize(angle: float) -> float:
    check_type(angle, float, "wrong type for angle")
    while angle < 0:
        angle += 2.0 * math.pi
    while angle > 2 * math.pi:
        angle -= 2.0 * math.pi
    return angle;


def angle_difference(a1: float, a2: float) -> float:
    a1 = normalize(a1)
    a2 = normalize(a2)
    if a1 > a2:
        dif = a1 - a2;
        if dif < math.pi:
            return dif, 1;
        else:
            return a2 + math.pi * 2.0 - a1, -1
    else:
        dif = a2 - a1;
        if dif < math.pi:
            return dif, -1
        else:
            return a1 + math.pi * 2.0 - a2, 1


def angle_between(value: float, lim1: float, lim2: float, inclusive: bool=False) -> bool:
    diff1, dir1 = angle_difference(value, lim1)
    diff2, dir2 = angle_difference(value, lim2)
    if inclusive:
        if diff1 == 0 or diff2 == 0:
            return True
    return (dir1 + dir2) == 0



from json_cpp import JsonObject, JsonList
from .coordinates import Coordinates, Coordinates_list
from .world import World


class Paths(JsonObject):
    def __init__(self):
        self.moves = Coordinates_list()
        self.steps = JsonList(list_type=int)
        JsonObject.__init__(self)




from .shape import *
from .location import *
from .coordinates import *
from .cell import *
from .world import *
from .display import *
from .util import *
from .experiment import *
from .agent_markers import *
from .graph import *
from .capture import *
from .visibility import *
from .frame import *
from .path import *
from .QuickBundles import *

